/* vim: set sw=8 ts=8 si et: */

/*
 * Author: Guido Socher (guido(at)linuxfocus.org). This program is distributed
 * under the terms of the Gnu Public License (GPL).
 */

/* Note: update the date/rev. in the abouttxt */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <gtk/gtk.h>
/* this should be a at least 100: */
#define LINELENGTH 220
#define DATABUFLEN 55
#ifndef GPPPCONFIGFILE
#define GPPPCONFIGFILE "/etc/ppp/gpppwrap.conf"
#endif
#ifndef GPPPOFF
#define GPPPOFF "/etc/ppp/scripts/ppp-off"
#endif
#ifndef GPPPON
#define GPPPON "/etc/ppp/scripts/ppp-on"
#endif

struct configdata{
	int phone_num_size; /* one elem in list means phone_num_size=1 */
	gchar *ppponarg[30]; /*max 30 phone numbers/peers scripts */
};
static gchar abouttxt[]=
"gpppwrap version 1.1, 23 May 2003 Written by\n\
Guido Socher.  This program is distributed under\n\
the terms of the GNU public license. gpppwrap\n\
provides a simple GUI wrapper around the\n\
powerful and flexible ppp-on/off scripts.\n\
It may be installed with a SUID ppp-on/off\n\
scripts to give ordinary users the possibility\n\
to execute the scripts. gpppwrap reads a\n\
list of different ISPs (option ppponarg) from\n\
/etc/ppp/gpppwrap.conf. gpppwrap calls the\n\
ppp-on script with the selected ppponarg\n\
option as argument.";

static gchar helptxt[]=
"Select an entry from the combo-box or change\n\
it as needed. A \"-\" starts a comment string.\n\
Press ppp-on to start the ppp-on dial out\n\
script. To terminate the connection just press\n\
ppp-off. This program does not check whether or\n\
not ppp connection is actually up. As it may take\n\
some time to dial in or terminate the connection\n\
you should not switch too fast between ppp-on\n\
and ppp-off.";

static int gppp_state=0; /*0 unknown, 1 ppp on, 2 ppp off*/

static GtkWidget *mw_txt_label;
static int ppp_was_on=0;

/*
 * Reads a line from file f, up to the size of the buffer.  The line in the
 * buffer will NOT include line termination, although any of (CR, LF, CRLF)
 * is accepted on input.  The return value is 0 on EOF,
 * 2 on out of buffer space and 1 if the line was terminated normally.
 * Tab and spaces are removed from the line.
 *
 * Passing in a buffer less than 2 characters long is not a terribly bright
 * idea.
 */
static int my_getline(char *buf, int buflen, FILE * f)
{
        int state=0; 
        int ignore_comment=0;
        char *p;
        int c;

        p = buf;

        /*lcount++; */
        while (1) {
                c = getc(f);
                if (c == '#') {
                        /*remove comments: put a \0 but continue reading
                         *to advance in the file to the end of the line*/
                        ignore_comment=1;
                        *p = 0;
                }
                if (c == '\n') {
                        *p = 0;
                        return 1;       /* Line terminated with \n or \r\n */
                }
                if (state) {
                        ungetc(c, f);
                        *p = 0;
                        return 1;       /* Line terminated with \r only */
                }
                if (c == EOF) {
                        *p = 0;
                        return (0);
                }
                if (c == '\r')
                        state = 1;
                else if (--buflen > 0 || ignore_comment) {
                        if (! ignore_comment) *p++ = c;
                } else {
                        ungetc(c, f);
                        *p = 0;
                        /*lcount--;*/
                        return 2;       /* Out of buffer space */
                }
        }
}       

/*
 * parse one line. The syntax is:
 * keyword:value
 * space is not significant
 *
 * return val --  keyword
 * 0              line not recognized
 * 1              ppponarg
 *
 * Comments start with # but were already removed by the my_getline() function.
 * The memory area that data points to is changed!
 * if return value > 0 then datawordbuf points the data (first non space char) 
 * character behind the keyword in the data buffer.
 * The buffer data is modified!!
 *
 * data is a normal 0 terminated string
 */
static int parse_line(char **datawordbuf,char *data){
        char *keystart;
        char *last_character_or_number_position;
        int after_colon=0;

        *datawordbuf=NULL;
        keystart=NULL;
        /*initialize: */
        last_character_or_number_position=data;
        while(*data){
                /* after the first ":" */
                if (*data==':' && after_colon==0){
                        *data=0;
                        data++;
                        after_colon=1;
                        continue;
                }
                if (isspace((int)*data)){
                        *data=' ';
                }
                if (!isspace((int)*data)){
                        /*remember the last non space character*/
                        last_character_or_number_position=data;
                }
                if (isgraph((int)*data)){
                        if (keystart && *datawordbuf==NULL && after_colon){
                                /* we do not break here as we have to
                                 * find the last isalnum character in
                                 * *datawordbuf */
                                *datawordbuf=data;
                        }
                        if (keystart==NULL && *datawordbuf==NULL){
                                keystart=data; 
                        }
                        data++;
                        continue;
                }
                /*we have e.g a line like this and are reading the
                 *space (or the colon) before *datawordbuf:
                 *ppponarg :  12 44 545 # tony's number*/
                if (*datawordbuf==NULL){
                        /*erase every non character */
                        *data=0;
                }
                data++;
        }
        if (*datawordbuf==NULL) return(0);

        /*remove tailing space after *datawordbuf:*/
        *(last_character_or_number_position+1)=0; 
        if (strcmp(keystart,"ppponarg")==0) return(1);
        return(0);
}
/*
 * read the config file and set the data in the configdata struct
 */
static void new_configdata(struct configdata *data){
        FILE *instream;
        char *datawordptr;
        int i=0;
        char buf[LINELENGTH];

        instream=fopen(GPPPCONFIGFILE,"r");
        if (instream==NULL){
                g_warning("gpppwrap: Can not read %s\n",GPPPCONFIGFILE);
                data->phone_num_size=0;
        }else{
                /* we make the buffer big enough so we don't need to check on
                 * my_getline()== 2 */
                while(my_getline(buf,LINELENGTH,instream)==1){
                        switch (parse_line(&datawordptr,buf)){
                                case 1: /*ppponarg*/
                                        data->ppponarg[i]=malloc(DATABUFLEN);
                                        if (data->ppponarg[i] == NULL) {
                                                g_error("gpppwrap: out of memory\n");
                                        }
                                        strncpy(data->ppponarg[i],datawordptr,DATABUFLEN-1);
                                        *(data->ppponarg[i] + DATABUFLEN)=0;
                                        i++;
                                        break;
                                /*we might get more key words in future*/
                        }

                }
                data->phone_num_size=i;
                fclose(instream);
        }
}
/*
 * get the phone number list from the config data object
 */
static GList *get_phone_numbers(struct configdata *data){
	static GList *list;
	int i=0;

	list= NULL;
	while(i < data->phone_num_size){
		list= g_list_append(list,data->ppponarg[i]);
		i++;
	}
        if (i==0){
                list= g_list_append(list,"");
        }
	return(list);
}
/*
 * destroy the popup window
 */
static gint ppp_popup_destroy(GtkWidget * widget, GtkWidget *data){
        if (data==NULL) return (TRUE); /*already dead*/
	gtk_widget_destroy (GTK_WIDGET (data));
	data = NULL;
	return (TRUE);
}
/*
 * show a dialog window a text label and an ok button to click 
 * the window away.
 * The text must contain already the necessary newline characters. 
 */
static void ppp_popup(GtkMenuItem *menuitem, gchar *text){
	GtkWidget *window;
	GtkWidget *button;
	GtkWidget *label;
	
	window = gtk_dialog_new ();
	/* the minimum size: Width, Hight */
	gtk_widget_set_usize( GTK_WIDGET (window), 330, 260);
        /*- button -*/
	button = gtk_button_new_with_label("Ok");
	/* expand, fill and padd by 10 off the border */
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area),
		button,TRUE, TRUE,10);
	gtk_signal_connect( GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(ppp_popup_destroy),(gpointer) window);
	gtk_widget_show (button);
        /*- label -*/
	label = gtk_label_new (text);
	gtk_box_pack_start(GTK_BOX (GTK_DIALOG (window)->vbox),label, TRUE,
		TRUE, 0);
	gtk_widget_show (label);
        /*- dialog is ready -*/
	gtk_widget_show (window);
}	
/*
 * Command must hold the absolute path to the executable.
 */
static int run_cmd(const char *command,const char *arg) {
        int pid;
        char *arglist[4];
        char cmdstr[100];
        char *env[]={"PATH=/sbin:/bin:/usr/sbin:/usr/bin",NULL}; 

        if (command == 0) return(1);
        if (strlen(command)+strlen(arg) > 90) return(1);
        if (access(command,F_OK)){  
                /*I don't want to text X_OK as setreuid is set in child only*/
                g_warning("gpppwrap: file %s does not exist\n",command);
                return(1);
        }
        strcpy(cmdstr,command);
        strcat(cmdstr," ");
        strcat(cmdstr,arg);
        strcat(cmdstr," &");
        pid = (int)fork();
        if (pid < 0) return(-1);
        if (pid == 0) {
                /*child*/
                /* no longer needed setreuid(0,0); ppp-on is now a suid wrapper */
                arglist[0] = "/bin/sh"; /*just sets the name for ps*/
                arglist[1] = "-c";
                arglist[2] = cmdstr;
                arglist[3] = 0;
                execve("/bin/sh", arglist, env);
                /*we get here only in case of error*/
                perror("gpppwrap"); 
        }
        /*parent*/
        /*wait for child to avoid zombie, should be fast as
         *we tell the shell to fork in background*/
        if (waitpid(pid, (int *)0, 0) != pid){
                g_warning("gpppwrap: waitpid failed\n");
                return(1);
        }
        return(0);
} 
static void setinfolabel(int texttemplate,char *ip){
        char *txt;
        char buffer[25];
        txt="error";
        switch (texttemplate){
                case 1: txt="  Last cmd: off   ";
                        break;
                case 2: txt="  Last cmd: on    ";
                        break;
                case 3: snprintf(buffer,21,"ip:%15s",ip);
                        txt=buffer;
                        break;
                case 4: txt="  ppp is off      ";
                        break;
        }
        gtk_label_set(GTK_LABEL(mw_txt_label),txt);
}
/*
 * the action associated with the ppp-off button
 */
static gint ppp_off_action(GtkWidget *widget, gpointer *ignore){
        /*check if we are already off*/
        setinfolabel(1,"");
        gppp_state=2; /*set state to off */
        run_cmd(GPPPOFF,"");
	return (TRUE);	
}
/*
 * the action associated with the ppp-on button
 */
static gint ppp_ck_ifconfig(gpointer data){
        int state=0; 
        int c,pppfound,i;
        char firststring[]="ppp";
        char secondstring[]="inet addr:";
        char ip[30];
        char ifconfig[]="/sbin/ifconfig";

        FILE *f;
        if (access(ifconfig,F_OK)){
                g_warning("gpppwrap: could not find %s\n",ifconfig);
                return (FALSE);	
        }
        f=popen(ifconfig,"r");
        if (!f){
                g_warning("gpppwrap: could not read data from ifconfig\n");
                return (FALSE);	
        }
        pppfound=0;
        i=0;
        /* read the output of ifconfig and find 3 times ppp at the
         * beginning of the line */
        while (1) {
                c = getc(f);
                if (c == EOF) {
                        break;
                }
                if (state==0){
                       if(firststring[i] && c == firststring[i]) {
                                i++;
                                if (firststring[i]=='\0'){
                                        /* match found */
                                        state++;
                                        i=0;
                                }
                       }else{
                               /* push back in case we have matched
                                * on a repeated character */
                               if (i == 1){
                                        ungetc(c, f);
                               }
                               i=0;
                       }
                       continue;
                }
                if (state==1){
                        /* search for second string */
                       if(secondstring[i] && c == secondstring[i]) {
                                i++;
                                if (secondstring[i]=='\0'){
                                        /* match found */
                                        state++;
                                        i=0;
                                        continue;
                                }
                       }else{
                               /* push back in case we have matched
                                * on a repeated character */
                               if (i == 1){
                                        ungetc(c, f);
                               }
                               i=0;
                       }
                }
                if (state==2){
                        /* read the ip address */
                        if (i < 20){
                                ip[i]=c;
                                i++;
                        }else{
                                state++;
                                break;
                        }
                        if (c == ' '){
                                state++;
                                break;
                        }
                }
        }
        pclose(f);
        ip[i]='\0';
        if (state==3){
                setinfolabel(3,ip);
                ppp_was_on=1;
        }else{
                /* do not set to off if it was never on */
                if (ppp_was_on){
                        setinfolabel(4,"");
                }
        }
	return (TRUE);	
}
/*
 * the action associated with the ppp-on button
 */
static gint ppp_on_action(GtkWidget *widget, GtkWidget *entry){
        /* there is no phone number longer than DATABUFLEN-1 digits */
	gchar *txt;
	gchar ph_num[DATABUFLEN];
	gchar *ph_num_ptr;
        int i=0;
	
        gppp_state=1; /*set state to on */
	txt=gtk_entry_get_text(GTK_ENTRY(entry));

	ph_num_ptr=ph_num;
        /* phone numbers may have a tailing comment starting with '-' */
	while(*txt && i < DATABUFLEN && *txt!='-'){
		if (isgraph((int)*txt)){
			*ph_num_ptr=*txt;
			ph_num_ptr++;
		}
		txt++;
                i++;
	}
        *ph_num_ptr='\0';
        setinfolabel(2,"");
	/*g_print("debug phone num to dial is %s\n",ph_num);*/
        run_cmd(GPPPON,ph_num);
	return (TRUE);	
}
/*
 * quit the whole application
 */
static void byebye(GtkWidget * widget, gpointer data){
	gtk_exit (0);
}

int main(int argc, char *argv[]){

	struct configdata confdat;
	struct configdata *confdat_ptr;
		
	GtkWidget *mainwindow;
	GtkWidget *the_file_menu,*file_menu;
	GtkWidget *the_help_menu,*help_menu;
	GtkWidget *menu_items,*menu_bar;
	GtkWidget *combobox;
        GtkEntry  *combobox_entry;
	GList     *combobox_items; 
	GtkWidget *button_on,*button_off;
	GtkWidget *vbox,*box1,*box2,*box3,*box4, *labelbox;

	/* parse GTK command line and take out gtk args. The rest is
         * still in argv[] */
	gtk_init(&argc, &argv);
	
	confdat_ptr=&confdat;
        new_configdata(confdat_ptr);

	mainwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        /* for some reason this does not work with my fvwm WM: */
        /*gtk_widget_set_uposition(GTK_WIDGET(mainwindow),10,110);*/
	/* minimum size */
	gtk_widget_set_usize( GTK_WIDGET(mainwindow), 300, 90);
	gtk_window_set_title( GTK_WINDOW(mainwindow), "gpppwrap");
	/* handle delete_event, usually by the 'close' option */
	gtk_signal_connect(GTK_OBJECT(mainwindow), "delete_event",
		GTK_SIGNAL_FUNC(byebye), NULL);
        gtk_signal_connect(GTK_OBJECT(mainwindow), "destroy",
                GTK_SIGNAL_FUNC(byebye), NULL);
	

	/* set up the boxes to do the packing: */
	vbox = gtk_vbox_new(FALSE,0); /* 0=no border */
	box1 = gtk_hbox_new(FALSE,1);
	box2 = gtk_hbox_new(FALSE,1);
	box3 = gtk_hbox_new(FALSE,1);
	box4 = gtk_hbox_new(FALSE,1);
	gtk_box_pack_start(GTK_BOX(vbox), box1, TRUE, TRUE, 0);
        /* put a little space between the 4 boxes: */
	gtk_box_pack_start(GTK_BOX(vbox), box2, TRUE, TRUE, 4);
	gtk_box_pack_start(GTK_BOX(vbox), box3, TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER (mainwindow), vbox);


	/* This is the file menu on the left. First we construct
         * the_file_menu which is the actual popup window part */
	the_file_menu=gtk_menu_new();
	menu_items= gtk_menu_item_new_with_label("Quit");
	gtk_menu_append(GTK_MENU(the_file_menu), menu_items);
	gtk_signal_connect_object(GTK_OBJECT(menu_items),"activate",
		GTK_SIGNAL_FUNC(byebye), NULL);
	/* Show the menu entry */
	gtk_widget_show(menu_items);

	/* lable of the menu */
	file_menu = gtk_menu_item_new_with_label("File");
	/* attach the_file_menu to the label */
	gtk_menu_item_set_submenu(GTK_MENU_ITEM (file_menu), the_file_menu);
	gtk_widget_show(file_menu);

	/* This is the help menu on the left. First we construct
	 * the_help_menu which is the actual popup window part */ 
	the_help_menu = gtk_menu_new();
	menu_items = gtk_menu_item_new_with_label("About");
	gtk_menu_append(GTK_MENU(the_help_menu), menu_items);
	gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
		GTK_SIGNAL_FUNC(ppp_popup), (gpointer) abouttxt);
	/* Show the menu entry */ 
	gtk_widget_show(menu_items);  
	menu_items = gtk_menu_item_new_with_label("Help");
	gtk_menu_append(GTK_MENU(the_help_menu), menu_items);
	gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
		GTK_SIGNAL_FUNC(ppp_popup), (gpointer) helptxt);
	/* Show the menu entry */ 
	gtk_widget_show(menu_items);
	
	/* lable of the menu */ 
	help_menu = gtk_menu_item_new_with_label("Help");
	/* attach the_help_menu to the label */ 
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_menu),the_help_menu);
	gtk_widget_show(help_menu);

	/* now come the actual menu bar. we must connect the menu label to it */
	menu_bar = gtk_menu_bar_new();
	gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), file_menu);
	gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), help_menu);
	gtk_menu_item_right_justify(GTK_MENU_ITEM (help_menu));

	/* pack: expand and fill no padd */
	gtk_box_pack_start(GTK_BOX(box1), menu_bar, TRUE, TRUE, 0);
	/* menu ready */
	gtk_widget_show(menu_bar);
	

	/* the text in the main window */
	mw_txt_label = gtk_label_new("  ppp-on/off  ");
        gtk_label_set_justify(GTK_LABEL(mw_txt_label),GTK_JUSTIFY_LEFT);
	labelbox = gtk_hbox_new(FALSE, 0);
	/* make the lable big enough by putting a box arround it
         * the hight does not matter as it is minimum size values: */
	gtk_widget_set_usize(GTK_WIDGET(labelbox),120,8);
	/* pack: expand, fill, padd 4*/  
	gtk_box_pack_start(GTK_BOX(box2), labelbox, TRUE, TRUE, 2);
        /* put the label into the box: */
	gtk_container_add(GTK_CONTAINER (labelbox), mw_txt_label);
	gtk_widget_show(mw_txt_label);
        gtk_widget_show(labelbox);


	/* a combo box to enter and select various phone numbers */
	/* first add all phone numbers to the list */
	combobox_items=get_phone_numbers(confdat_ptr);
	combobox = gtk_combo_new(); 
	gtk_combo_set_popdown_strings(GTK_COMBO(combobox),combobox_items); 
	/* pack: expand, no fill, padd 5 */
	gtk_box_pack_start(GTK_BOX(box2), combobox, TRUE, TRUE, 4);
	combobox_entry=GTK_ENTRY(GTK_COMBO(combobox)->entry);
	gtk_widget_show(combobox); 

	
	/* the status text */
	/* mw_txt_status_label = gtk_label_new("status: ");
         * gtk_label_set_justify(GTK_LABEL(mw_txt_status_label),GTK_JUSTIFY_LEFT);
	 * gtk_box_pack_start(GTK_BOX(box3),mw_txt_status_label,FALSE,FALSE,10);
	 * gtk_widget_show(mw_txt_status_label);
         */
	
	/* now comes the next row with the ppp-on/off buttons */
	button_on=gtk_button_new_with_label("run ppp-on");
	gtk_signal_connect(GTK_OBJECT(button_on), "clicked",
		GTK_SIGNAL_FUNC(ppp_on_action),(gpointer) combobox_entry);
	/* pack: expand, fill, padd 1 */
	gtk_box_pack_start(GTK_BOX(box3), button_on,TRUE, TRUE,0);
	gtk_widget_show(button_on);
	
	button_off=gtk_button_new_with_label("run ppp-off");
	gtk_signal_connect(GTK_OBJECT(button_off), "clicked",
		GTK_SIGNAL_FUNC(ppp_off_action),NULL);
	/* pack: expand, fill, padd 1 */
	gtk_box_pack_start(GTK_BOX(box3), button_off,TRUE, TRUE,0);
	gtk_widget_show(button_off);
	
        /* every 8 seconds we check if the interface ppp is up*/
        gtk_timeout_add( 8000,ppp_ck_ifconfig, NULL);
	

	/* yeaaaaa GUI is ready */
	gtk_widget_show(box1);
	gtk_widget_show(box2);
	gtk_widget_show(box3);
	gtk_widget_show(vbox);
	gtk_widget_show(mainwindow);

	/* go to main loop */
	gtk_main();
	return 0;
}

